#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2023 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tools for profiling scaling alternatives.
"""
import sys

from pyomo.environ import check_optimal_termination, Constraint
from pyomo.common.tempfiles import TempfileManager

from idaes.core.util.scaling import jacobian_cond
from idaes.core.scaling.autoscaling import AutoScaler
from idaes.core.scaling.custom_scaler_base import (
    CustomScalerBase,
    ConstraintScalingScheme,
)
from idaes.core.solvers import get_solver


class ScalingProfiler:
    """
    Class for running a set of constraint scaling methods on a model and reporting the
    effect on model condition number and solver behavior.

    Users should call the profile_scaling_methods method to generate a dict of results or
    the report_scaling_profiles method for a stream-based output.

    Users are expected to provide callback functions to 1) construct an initialized model
    that will be used for profiling, 2) apply user-defined variable scaling (used for the
    imperfect information case) and 3) perturb the model from the initialized state to
    test how well the model solves (optional).

    Users may also provide a dict of scaling methods they wish to apply using the scaling_methods
    argument. If this is not provided, the tool will default to applying all the scaling methods
    defined by the AutoScaler and CustomScalerBase classes.

    **NOTE** methods from the AutoScaler class are applied to Pyomo Blocks, whilst those from
    CustomScalerBase are applied to individual ConstraintDatas. The profiling tool assumes that
    methods will be applied to ConstraintDatas unless the `block_based` keyword argument is set to True
    for the scaling method.

    Args:
        build_model: callback to use to construct initialized model for testing
        user_scaling: callback to use to apply user-defined scaling to initialized model
        perturb_states: (optional) callback to use to perturb model state for re-solve tests
        scaling_methods: (optional) dict of constraint scaling methods to profile. {"Name": (method, kwargs)}
        solver: (optional) Pyomo solver object to use for re-solve tests
    """

    def __init__(
        self,
        build_model,
        user_scaling,
        perturb_state=None,
        scaling_methods: dict = None,
        solver=None,
    ):
        """
        Sets up a framework for applying different scaling methods to a model and compiling a
        report of their effects on the Jacobian condition number and how easily the scaled
        model can be solved for a perturbed state.
        """
        self._build_model = build_model
        self._user_scaling = user_scaling
        self._perturb_state = perturb_state
        self._scaling_methods = scaling_methods
        self._solver = solver

        if self._solver is None:
            self._solver = get_solver("ipopt_v2", writer_config={"scale_model": True})

        if self._scaling_methods is None:
            ascaler = AutoScaler()
            cscaler = CustomScalerBase()

            self._scaling_methods = {
                "Vars Only": (None, {}),
                "Harmonic": (
                    cscaler.scale_constraint_by_nominal_value,
                    {"scheme": ConstraintScalingScheme.harmonicMean},
                ),
                "Inverse Sum": (
                    cscaler.scale_constraint_by_nominal_value,
                    {"scheme": ConstraintScalingScheme.inverseSum},
                ),
                "Inverse Root Sum Squares": (
                    cscaler.scale_constraint_by_nominal_value,
                    {"scheme": ConstraintScalingScheme.inverseRSS},
                ),
                "Inverse Maximum": (
                    cscaler.scale_constraint_by_nominal_value,
                    {"scheme": ConstraintScalingScheme.inverseMaximum},
                ),
                "Inverse Minimum": (
                    cscaler.scale_constraint_by_nominal_value,
                    {"scheme": ConstraintScalingScheme.inverseMinimum},
                ),
                "Nominal L1 Norm": (
                    cscaler.scale_constraint_by_nominal_derivative_norm,
                    {"norm": 1},
                ),
                "Nominal L2 Norm": (
                    cscaler.scale_constraint_by_nominal_derivative_norm,
                    {"norm": 2},
                ),
                "Actual L1 Norm": (
                    ascaler.scale_constraints_by_jacobian_norm,
                    {"norm": 1, "block_based": True},
                ),
                "Actual L2 Norm": (
                    ascaler.scale_constraints_by_jacobian_norm,
                    {"norm": 2, "block_based": True},
                ),
            }

    def profile_scaling_methods(self):
        """
        Generate results for all provided scaling methods.

        For each scaling method, calculate the Jacobian condition number and re-solve
        the model with both user-provided variable scaling and perfect variable scaling
        (scaling by inverse magnitude). A base case with no scaling applied is also
        run for reference.

        Args:
            None

        Returns:
            dict with results for all scaling methods
        """
        # Generate data for unscaled model
        m = self._build_model()
        unscaled = jacobian_cond(m, scaled=False)
        stats = self._solved_perturbed_state(m)

        results = {
            "Unscaled": {"Manual": {"condition_number": unscaled, **stats}},
        }

        # Run other cases
        for case, (meth, margs) in self._scaling_methods.items():
            results[case] = self.run_case(meth, **margs)

        return results

    def report_scaling_profiles(self, stream=None):
        """
        Run scaling profile workflow nad report results to a stream.

        Args:
            stream: StringIO object to write result to (default=stdout)

        Returns:
            None
        """
        results = self.profile_scaling_methods()

        self.write_profile_report(results, stream)

    def write_profile_report(self, results: dict, stream=None):
        """
        Write a report on the comparison of scaling methods to a stream based on
        existing results dict.

        Args:
            results: dict containing results from a scaler profiling run
            stream: StringIO object to write result to (default=stdout)

        Returns:
            None
        """
        # If stream is None, default to stdout
        if stream is None:
            stream = sys.stdout

        # Get length of longest method name for padding
        max_str = max([len(i) for i in results.keys()])
        if max_str < len("Scaling Method"):
            max_str = len("Scaling Method")
        # Length of each stats field is 22 characters, plus 4 for column dividers
        # Max line length is thus longest string name + 2 columns of 22+4 characters
        max_line = max_str + 26 * 2

        # Write header rows
        stream.write(f"\n{'='*max_line}\n")
        stream.write("Scaling Profile Report\n")
        stream.write(f"{'-' * max_line}\n")
        # Pad User Scaling columns to full column width (22)
        stream.write(
            f"{'Scaling Method': <{max_str}} || {'User Scaling': <{22}} || Perfect Scaling\n"
        )

        # Iterate over keys in results and write summary for each scaling method
        for k, v in results.items():
            if v["Manual"]["solved"]:
                msolved = "Solved"
            else:
                msolved = "Failed"
            # Pad iterations to 3 characters - hopefully we don;t see more than 999 iterations
            miters = f"{v['Manual']['iterations']: <{3}}"
            stream.write(
                f"{k: <{max_str}} || {v['Manual']['condition_number']:.3E} | {msolved} {miters} "
            )

            if "Auto" in v.keys():
                if v["Auto"]["solved"]:
                    asolved = "Solved"
                else:
                    asolved = "Failed"
                # Pad iterations again
                aiters = f"{v['Auto']['iterations']: <{3}}"
                stream.write(
                    f"|| {v['Auto']['condition_number']:.3E} | {asolved} {aiters}\n"
                )
            else:
                # Add training column divider but no auto column
                stream.write("||\n")

        # Write footer row
        stream.write(f"{'=' * max_line}\n")

    def run_case(self, scaling_method, **kwargs):
        """
        Run case for a given scaling method with both perfect and imperfect scaling information.

        Args:
            scaling_method: constraint scaling method to be tested
            kwargs: keyword argument to be passed to scaling method

        Returns:
            dict summarising results of scaling case
        """
        block_based = kwargs.pop("block_based", False)

        # Imperfect information
        manual = self._run_scenario(
            scaling_method, perfect=False, block_based=block_based, **kwargs
        )

        # Perfect information
        perfect = self._run_scenario(
            scaling_method, perfect=True, block_based=block_based, **kwargs
        )

        return {"Manual": manual, "Auto": perfect}

    def _scale_vars(self, model, perfect=False):
        """
        Apply variable scaling to model.
        """
        if perfect:
            scaler = AutoScaler()
            scaler.scale_variables_by_magnitude(model)
            return model

        self._user_scaling(model)

        return model

    def _apply_scaling(self, model, scaling_method, block_based, **kwargs):
        """
        Collect stats for a given scaling method.
        """
        if scaling_method is not None:
            if block_based:
                scaling_method(model, **kwargs)
            else:
                for c in model.component_data_objects(
                    ctype=Constraint, descend_into=True
                ):
                    scaling_method(c, **kwargs)

    def _run_scenario(self, scaling_method, block_based, perfect, **kwargs):
        """
        Run a single scenario
        """
        m = self._build_model()
        self._scale_vars(m, perfect=perfect)
        self._apply_scaling(m, scaling_method, block_based=block_based, **kwargs)

        cond = jacobian_cond(m, scaled=True)
        stats = self._solved_perturbed_state(m)

        return {"condition_number": cond, **stats}

    def _solved_perturbed_state(self, model):
        """
        Run re-solve tests if perturb_state callback provided.
        """
        if self._perturb_state is None:
            return {}

        self._perturb_state(model)

        TempfileManager.push()
        tempfile = TempfileManager.create_tempfile(suffix="ipopt_out", text=True)
        opts = {"output_file": tempfile}

        try:
            status_obj = self._solver.solve(model, options=opts, tee=True)
            solved = True
            if not check_optimal_termination(status_obj):
                solved = False

            iters, iters_in_restoration, iters_w_regularization = (
                self._parse_ipopt_output(tempfile)
            )

            return {
                "solved": solved,
                "termination_message": status_obj.solver.termination_message,
                "iterations": iters,
                "iters_in_restoration": iters_in_restoration,
                "iters_w_regularization": iters_w_regularization,
            }
        except RuntimeError as err:
            # Likely a critical solver failure
            return {
                "solved": False,
                "termination_message": str(err),
                "iterations": -1,
                "iters_in_restoration": -1,
                "iters_w_regularization": -1,
            }

    def _parse_ipopt_output(self, ipopt_file):
        """
        Parse an IPOPT output file and return:

        * number of iterations
        * time in IPOPT

        Returns
        -------
           Returns a tuple with (solve status object, bool (solve successful or
           not), number of iters, solve time)
        """
        # ToDO: Check for final iteration with regularization or restoration

        iters = 0
        iters_in_restoration = 0
        iters_w_regularization = 0
        # parse the output file to get the iteration count, solver times, etc.
        with open(ipopt_file, "r") as f:
            parseline = False
            for line in f:
                if line.startswith("iter"):
                    # This marks the start of the iteration logging, set parseline True
                    parseline = True
                elif line.startswith("Number of Iterations....:"):
                    # Marks end of iteration logging, set parseline False
                    parseline = False
                    tokens = line.split()
                    iters = int(tokens[3])
                elif parseline:
                    # Line contains details of an iteration, look for restoration or regularization
                    tokens = line.split()
                    try:
                        if not tokens[6] == "-":
                            # Iteration with regularization
                            iters_w_regularization += 1
                        if tokens[0].endswith("r"):
                            # Iteration in restoration
                            iters_in_restoration += 1
                    except IndexError:
                        # Blank line at end of iteration list, so assume we hit this
                        pass

        return iters, iters_in_restoration, iters_w_regularization



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
from copy import deepcopy
from ..util.util import myPointEq


def areMotifViaTransF(D1, D2, TransF, blnPreserveIndexing=False, blnIgnoreVoid=True):
    """

    Args:
        D1: param D2:
        TransF: param blnPreserveIndexing:  (Default value = False)
        blnIgnoreVoid: Default value = True)
        D2:
        blnPreserveIndexing:  (Default value = False)

    Returns:

    """
    return D1.isEquivalentTo(
        D2.getTransformed(TransF),
        blnPreserveIndexing=blnPreserveIndexing,
        blnIgnoreVoid=blnIgnoreVoid,
    )


def areMotifViaTransFs(D1, D2, TransFs, blnPreserveIndexing=False, blnIgnoreVoid=True):
    """

    Args:
        D1: param D2:
        TransFs: param blnPreserveIndexing:  (Default value = False)
        blnIgnoreVoid: Default value = True)
        D2:
        blnPreserveIndexing:  (Default value = False)

    Returns:

    """
    for TransF in TransFs:
        if areMotifViaTransF(
            D1,
            D2,
            TransF,
            blnPreserveIndexing=blnPreserveIndexing,
            blnIgnoreVoid=blnIgnoreVoid,
        ):
            return True
    return False


def getEnumConfs(
    Motifs, TransFs, MotifToConfMap=None, blnPreserveMotifLocs=True, DBL_TOL=1e-5
):
    """

    Args:
        DBL_TOL:
        blnPreserveMotifLocs:
        Motifs: param TransFs:
        MotifToConfMap: Default value = None)
        TransFs:

    Returns:

    """
    if MotifToConfMap is not None:
        # del MotifToConfMap[:]
        MotifToConfMap.clear()
    result = []
    for iMotif, Motif in enumerate(Motifs):
        if MotifToConfMap is not None:
            MotifToConfMap.append([len(result)])
        result.append(Motif)
        for TransF in TransFs:
            Conf = Motif.getTransformed(TransF)
            blnUniqueConf = True
            for r in result:
                if Conf.isEquivalentTo(r):
                    blnUniqueConf = False
                    break
            if blnUniqueConf:
                if MotifToConfMap is not None:
                    MotifToConfMap[iMotif].append(len(result))
                # NOTE: This section is important if we want to just assume
                #       that each index is the same location.
                #       This is important for use in models.
                if blnPreserveMotifLocs:
                    ConfCopy = Conf
                    Conf = deepcopy(Motif)  # So they have the same points<->indexes
                    for l in range(len(Motif)):
                        for l2 in range(len(ConfCopy)):
                            if myPointEq(
                                Motif.Canvas.Points[l],
                                ConfCopy.Canvas.Points[l2],
                                DBL_TOL,
                            ):
                                Conf.setContent(l, ConfCopy.Contents[l2])
                                break
                result.append(Conf)
    return result



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################

__author__ = "John Eslick"

import pytest
from idaes.models.properties.general_helmholtz import (
    HelmholtzParameterBlock,
    PhaseType,
    StateVars,
    AmountBasis,
)
from idaes.models.properties.tests.test_harness import PropertyTestHarness
from idaes.models.properties.general_helmholtz import helmholtz_available


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMix(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {"pure_component": "h2o", "phase_presentation": PhaseType.MIX}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {"pure_component": "h2o", "phase_presentation": PhaseType.LG}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicL(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {"pure_component": "h2o", "phase_presentation": PhaseType.L}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicV(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {"pure_component": "h2o", "phase_presentation": PhaseType.G}
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMixSP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "state_vars": StateVars.PS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLVSP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "state_vars": StateVars.PS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLSP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "state_vars": StateVars.PS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicVSP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "state_vars": StateVars.PS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMixUP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "state_vars": StateVars.PU,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLVUP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "state_vars": StateVars.PU,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLUP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "state_vars": StateVars.PU,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicVUP(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "state_vars": StateVars.PU,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMixMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLVMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicVMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMixSPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "state_vars": StateVars.PS,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLVSPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "state_vars": StateVars.PS,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLSPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "state_vars": StateVars.PS,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicVSPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "state_vars": StateVars.PS,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicMixUPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "state_vars": StateVars.PU,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLVUPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "state_vars": StateVars.PU,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicLUPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "state_vars": StateVars.PU,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


@pytest.mark.skipif(not helmholtz_available(), reason="General Helmholtz not available")
class TestBasicVUPMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "state_vars": StateVars.PU,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


class TestBasicMixTPXMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.MIX,
            "state_vars": StateVars.TPX,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


class TestBasicLVTPXMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.LG,
            "state_vars": StateVars.TPX,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


class TestBasicLTPXMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.L,
            "state_vars": StateVars.TPX,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True


class TestBasicVTPXMass(PropertyTestHarness):
    def configure(self):
        self.prop_pack = HelmholtzParameterBlock
        self.param_args = {
            "pure_component": "h2o",
            "phase_presentation": PhaseType.G,
            "state_vars": StateVars.TPX,
            "amount_basis": AmountBasis.MASS,
        }
        self.prop_args = {}
        self.has_density_terms = True
        # Helmholtz package initialization has no solver calls, so can't fail
        self.skip_initialization_raises_exception_test = True



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for methods from Perry's

All methods and parameters from:

Perry's Chemical Engineers' Handbook, 7th Edition
Perry, Green, Maloney, 1997, McGraw-Hill

All parameter indices based on conventions used by the source

Authors: Andrew Lee
"""

import pytest
import types

from pyomo.environ import ConcreteModel, Block, value, Var, Param, units as pyunits
from pyomo.common.config import ConfigBlock
from pyomo.util.check_units import assert_units_equivalent

from idaes.models.properties.modular_properties.pure.Perrys import *
from idaes.core.util.misc import add_object_reference
from idaes.core.base.property_meta import PropertyClassMetadata, UnitSet
from idaes.core.util.exceptions import ConfigurationError


@pytest.fixture()
def frame():
    m = ConcreteModel()

    # Create a dummy parameter block
    m.params = Block()

    m.params.config = ConfigBlock(implicit=True)
    m.params.config.parameter_data = {
        "dens_mol_liq_comp_coeff": {
            "eqn_type": 1,
            "1": 5.459,
            "2": 0.30542,
            "3": 647.13,
            "4": 0.081,
        },
        "cp_mol_liq_comp_coeff": {
            "1": 2.7637e05,
            "2": -2.0901e03,
            "3": 8.1250e00,
            "4": -1.4116e-2,
            "5": 9.3701e-06,
        },
        "enth_mol_form_liq_comp_ref": -285.83e3,
        "entr_mol_form_liq_comp_ref": 69.95,
    }
    m.params.config.include_enthalpy_of_formation = True

    # Also need to dummy configblock on the model for the test
    m.config = ConfigBlock(implicit=True)
    m.config.include_enthalpy_of_formation = True

    m.meta_object = PropertyClassMetadata()
    m.meta_object._default_units.set_units(
        temperature=pyunits.K,
        mass=pyunits.kg,
        length=pyunits.m,
        time=pyunits.s,
        amount=pyunits.mol,
    )

    def get_metadata(self):
        return m.meta_object

    m.get_metadata = types.MethodType(get_metadata, m)
    m.params.get_metadata = types.MethodType(get_metadata, m.params)

    # Add necessary parameters to parameter block
    m.params.temperature_ref = Var(initialize=273.16, units=pyunits.K)

    # Create a dummy state block
    m.props = Block([1])
    add_object_reference(m.props[1], "params", m.params)

    m.props[1].temperature = Var(initialize=273.16, units=pyunits.K)

    return m


@pytest.mark.unit
def test_cp_mol_liq_comp(frame):
    Perrys.cp_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.cp_mol_liq_comp_coeff_1, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_1) == 2.7637e05
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_2, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_2) == -2.0901e03
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_3, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_3) == 8.1250e00
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_4, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_4) == -1.4116e-2
    assert isinstance(frame.params.cp_mol_liq_comp_coeff_5, Var)
    assert value(frame.params.cp_mol_liq_comp_coeff_5) == 9.3701e-06

    expr = Perrys.cp_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(76.150, rel=1e-3)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(89.390, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_enth_mol_liq_comp(frame):
    Perrys.enth_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.enth_mol_form_liq_comp_ref, Var)
    assert value(frame.params.enth_mol_form_liq_comp_ref) == -285.83e3

    expr = Perrys.enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.enth_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(-265423, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_enth_mol_liq_comp_no_formation(frame):
    frame.config.include_enthalpy_of_formation = False
    frame.params.config.include_enthalpy_of_formation = False

    Perrys.enth_mol_liq_comp.build_parameters(frame.params)

    assert not hasattr(frame.params, "enth_mol_form_liq_comp_ref")

    expr = Perrys.enth_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == 0

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(-265423 + 285830, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol)


@pytest.mark.unit
def test_entr_mol_liq_comp(frame):
    Perrys.entr_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.entr_mol_form_liq_comp_ref, Var)
    assert value(frame.params.entr_mol_form_liq_comp_ref) == 69.95

    expr = Perrys.entr_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == value(frame.params.entr_mol_form_liq_comp_ref)

    frame.props[1].temperature.value = 533.15
    assert value(expr) == pytest.approx(122.05, rel=1e-3)

    assert_units_equivalent(expr, pyunits.J / pyunits.mol / pyunits.K)


@pytest.mark.unit
def test_dens_mol_liq_comp(frame):
    Perrys.dens_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff_eqn_type, Param)
    assert value(frame.params.dens_mol_liq_comp_coeff_eqn_type) == 1

    expr = Perrys.dens_mol_liq_comp.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(55.583e3, rel=1e-4)

    frame.props[1].temperature.value = 333.15
    assert value(expr) == pytest.approx(54.703e3, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol / pyunits.m**3)


@pytest.mark.unit
def test_dens_mol_liq_comp_eqn_1(frame):
    Perrys.dens_mol_liq_comp_eqn_1.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff_1, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_1) == 5.459
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_2, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_2) == 0.30542
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_3, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_3) == 647.13
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_4, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_4) == 0.081

    expr = Perrys.dens_mol_liq_comp_eqn_1.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(55.583e3, rel=1e-4)

    frame.props[1].temperature.value = 333.15
    assert value(expr) == pytest.approx(54.703e3, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol / pyunits.m**3)


@pytest.mark.unit
def test_dens_mol_liq_comp_eqn_2(frame):
    Perrys.dens_mol_liq_comp_eqn_2.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff_1, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_1) == 5.459
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_2, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_2) == 0.30542
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_3, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_3) == 647.13
    assert isinstance(frame.params.dens_mol_liq_comp_coeff_4, Var)
    assert value(frame.params.dens_mol_liq_comp_coeff_4) == 0.081

    expr = Perrys.dens_mol_liq_comp_eqn_2.return_expression(
        frame.props[1], frame.params, frame.props[1].temperature
    )
    assert value(expr) == pytest.approx(4.99375e10, rel=1e-4)

    frame.props[1].temperature.value = 333.15
    assert value(expr) == pytest.approx(7.48194e10, rel=1e-4)

    assert_units_equivalent(expr, pyunits.mol / pyunits.m**3)


@pytest.mark.unit
def test_dens_mol_liq_comp_exceptions(frame):
    frame.params.config.parameter_data["dens_mol_liq_comp_coeff"]["eqn_type"] = 3
    with pytest.raises(ConfigurationError):
        Perrys.dens_mol_liq_comp.build_parameters(frame.params)

    assert isinstance(frame.params.dens_mol_liq_comp_coeff_eqn_type, Param)
    assert value(frame.params.dens_mol_liq_comp_coeff_eqn_type) == 3

    with pytest.raises(ConfigurationError):
        Perrys.dens_mol_liq_comp.return_expression(
            frame.props[1], frame.params, frame.props[1].temperature
        )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""Commandline interface for convergence testing tools"""
# This code is deprecated
# pylint: disable=missing-function-docstring

__author__ = "John Eslick"

import click
from pyomo.common.dependencies import attempt_import
from idaes.commands import cb

cnv = attempt_import("idaes.core.util.convergence.convergence_base")[0]


@cb.command(name="convergence-sample", help="Create a convergence sample file.")
@click.option(
    "-e",
    "--evaluation-class",
    default=None,
    type=str,
    required=True,
    help="Convergence evaluation class",
)
@click.option(
    "-s",
    "--sample-file",
    default=None,
    type=str,
    required=True,
    help="Output sample file",
)
@click.option(
    "-N",
    "--number-samples",
    default=None,
    type=int,
    required=True,
    help="Number of samples",
)
@click.option("--seed", default=None, type=int, help="Random number generator seed")
@click.option(
    "-m",
    "--convergence_module",
    default=None,
    type=str,
    required=False,
    help="Additional module that registers ConvergenceEvaluation classes",
)
def convergence_sample(
    evaluation_class, sample_file, number_samples, seed, convergence_module
):
    click.echo(
        "The command line interface for convergence testing has been deprecated and no longer works. "
        "A new API has been added to run convergence tests directly from the ConvergenceEvaluation object."
    )


@cb.command(name="convergence-eval", help="Run convergence sample evaluation.")
@click.option(
    "-s",
    "--sample-file",
    default=None,
    type=str,
    required=True,
    help="Path of sample file to run",
)
@click.option(
    "-D",
    "--dmf",
    default=None,
    type=str,
    required=False,
    help="Path to Data Management Framework (DMF) workspace",
)
@click.option(
    "-r",
    "--report-file",
    default=None,
    type=str,
    required=False,
    help="Optional text report file path",
)
@click.option(
    "-j",
    "--json-file",
    default=None,
    type=str,
    required=False,
    help="Optional json file to save results and stats",
)
@click.option(
    "-m",
    "--convergence_module",
    default=None,
    type=str,
    required=False,
    help="Additional module that registers ConvergenceEvaluation classes",
)
@click.option(
    "--single-sample",
    default=None,
    type=str,
    help="Run only a single sample with given name",
)
def convergence_eval(
    sample_file, dmf, report_file, json_file, convergence_module, single_sample
):
    click.echo(
        "The command line interface for convergence testing has been deprecated and no longer works. "
        "A new API has been added to run convergence tests directly from the ConvergenceEvaluation object."
    )


@cb.command(name="convergence-search", help="Search for convergence test classes.")
@click.option(
    "-r",
    "--regex",
    default=None,
    type=str,
    help="Optional regular expression to filter registered convergence classes",
)
@click.option(
    "-m",
    "--convergence_module",
    default=None,
    type=str,
    required=False,
    help="Optional additional module that registers convergence classes",
)
def convergence_search(regex, convergence_module):
    click.echo(
        "The command line interface for convergence testing has been deprecated and no longer works. "
        "A new API has been added to run convergence tests directly from the ConvergenceEvaluation object."
    )



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Wilke rules to obtain mixture viscosity of a low pressure gas from pure component
viscosities given by visc_d_phase_comp. See Section 9-5-2 of The Properties of
Gases and Liquids, 5th Ed., for a detailed description of the method. Table 9-4
shows this method producing mixture viscosity within 5% of true viscosity, with
the notable exception of hydrogen-containing systems which have some errors up
to 12%. Those error figures are using experimentally-determined pure component
viscosity; additional error can accumulate when those viscosities must be
estimated.
"""
import pyomo.environ as pyo


class ViscosityWilke(object):
    """Class to implement Wilke method to calculate mixture dynamic viscosity"""

    @staticmethod
    def build_parameters(pobj):
        """
        Ensures there's a method to build phi_ij associated with the phase. The problem: the
        transport_properties_options dictionary may not exist, and, if it does exist, it may not have
        an explicit method assigned for the callback. Therefore create the dictionary if it doesn't exist
        and ensure the correct field is populated with the default callback.
        """
        # Have to use get method because transport_properties_options dictionary may not exist
        if (
            pobj.config.transport_property_options.get("viscosity_phi_ij_callback")
            is None
        ):
            pobj.config.transport_property_options["viscosity_phi_ij_callback"] = (
                wilke_phi_ij_callback
            )

    @staticmethod
    def build_phi_ij(b, p):
        """Method to build mixing function phi_ij"""
        pobj = b.params.get_phase(p)

        if not hasattr(b, "_visc_d_phase_comp"):
            b._make_visc_d_phase_comp()  # pylint: disable=protected-access

        if not hasattr(b, "visc_d_phi_ij"):
            mw_dict = {
                k: b.params.get_component(k).mw for k in b.components_in_phase(p)
            }

            def phi_rule(blk, i, j):
                return pobj.config.transport_property_options[
                    "viscosity_phi_ij_callback"
                ](blk, i, j, p, mw_dict)

            b.visc_d_phi_ij = pyo.Expression(
                b.components_in_phase(p),
                b.components_in_phase(p),
                rule=phi_rule,
                doc="Intermediate quantity for calculating gas mixture viscosity and thermal conductivity",
            )

    class visc_d_phase(object):
        """Method to construct gas mixture dynamic viscosity"""

        @staticmethod
        def build_parameters(pobj):
            """Build parameters"""
            ViscosityWilke.build_parameters(pobj)

        @staticmethod
        def return_expression(b, p):
            """Return expression for visc_d_phase"""
            # Properties of Gases and Liquids, Eq. 9-5.14
            ViscosityWilke.build_phi_ij(b, p)

            return sum(
                [
                    b.mole_frac_phase_comp[p, i]
                    * b._visc_d_phase_comp[p, i]  # pylint: disable=protected-access
                    / sum(
                        [
                            b.mole_frac_phase_comp[p, j] * b.visc_d_phi_ij[i, j]
                            for j in b.components_in_phase(p)
                        ]
                    )
                    for i in b.components_in_phase(p)
                ]
            )


def wilke_phi_ij_callback(b, i, j, pname, mw_dict):
    """Equation 9.5.14 from Properties of Gases and Liquids.

    .. math::

      \\phi_{ij} = \\frac{[1 + (\\mu_i/\\mu_j)^{1/2}(M_j/M_i)^{1/4}]^2}{[8(1+M_i/M_j)]^{1/2}}

    in which :math:`M_i` is the molecular weight of component :math:`i`.
    """
    visc_i = b._visc_d_phase_comp[pname, i]  # pylint: disable=protected-access
    visc_j = b._visc_d_phase_comp[pname, j]  # pylint: disable=protected-access
    return (
        1 + pyo.sqrt(visc_i / visc_j) * (mw_dict[j] / mw_dict[i]) ** 0.25
    ) ** 2 / pyo.sqrt(8 * (1 + mw_dict[i] / mw_dict[j]))


def herring_zimmer_phi_ij_callback(b, i, j, pname, mw_dict):
    """Equation 9.5.17 from Properties of Gases and Liquids.

    .. math::

      \\phi_{ij} = \\left(\\frac{M_j}{M_i}\\right)^{1/2}

    in which :math:`M_i` is the molecular weight of component :math:`i`.
    """
    return pyo.sqrt(mw_dict[j] / mw_dict[i])



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Main Assumptions:

- Fireside boiler surrogate model with mass and energy balance enforced
- Flue gas compositions of major species O2, N2, CO2, H2O and SO2 are
  calculated based on elemental mass balance
- Flue gas composition of NO is calculated based on algebraic surrogate model
- Unburned organic elements are calculated based on algebraic surrogate model
- Heat duties in waterwall zones and platen SH and roof are calculated based
  on surrogate model functions
- Flue Gas Exit Temperature (FGET) is calculated based on energy balance

The model has two inlets:

* primary air inlet: PA, flow rate and composition

* secondary air inlet: SA, flow rate and composition

* primary_air_moist: primary air minus vaporized moisture (calculated)


The model has two outlets:

* Flue gas (gas only): flow rate and composition calculated by mass balance
* flyash: Fly ash with unburned organic elements (solid phase only)


Notes:

1. SR versus raw coal flow rate can be specified by a flowsheet constraint
   and should be consistent with the surrogate model.
2. PA to raw coal flow rate ratio can be specified by a flowsheet constraint
3. Fraction of coal moisture vaporized in mill should be consistent with
   the surrogate model (typically 0.6)
4. Moisture mass fraction in raw coal is a user input and should be consistent
   with the values used to train the surrogate model


Surrogate models need to be provided by the user:

* heat duty, one for each water wall zone
* 1 for platen super heater
* 1 for roof and backpass
* 1 for NOx prediction
* 1 for flyash

Note that the user can provide either surrogate models or fixed numeric values

The surrogate models are typically a function of:

* raw coal mass flow rate,
* moisture mass fraction of raw coal,
* overall stoichiometric ratio,
* lower furnace Stoichiometric ratio,
* Secondary air temperature

"""
# Import Pyomo libraries
from pyomo.common.config import ConfigBlock, ConfigValue, Bool

# TODO: log is reauired for eval - need to work out how to make this explicit
# pylint: disable=W0611
from pyomo.environ import (
    Var,
    Param,
    exp,
    RangeSet,
    Constraint,
    log,
)

# Import IDAES cores
from idaes.core import declare_process_block_class, UnitModelBlockData, useDefault

from idaes.core.util.model_statistics import degrees_of_freedom

from idaes.core.util.config import is_physical_parameter_block, DefaultBool
from idaes.core.util.exceptions import ConfigurationError
import idaes.logger as idaeslog
import idaes.core.util.scaling as iscale
from idaes.core.util.constants import Constants as const
from idaes.core.solvers import get_solver

__author__ = "Boiler Team (J. Ma, M. Zamarripa)"
__version__ = "1.0.0"


# ----------------------------------------------------------------------------
@declare_process_block_class("BoilerFireside")
class BoilerFiresideData(UnitModelBlockData):
    """
    Boiler fire-side surrogate model with enforced mass and energy balance
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=DefaultBool,
            default=useDefault,
            description="Dynamic model flag",
            doc="""Indicates whether this model will be dynamic or not,
**default** = useDefault.
**Valid values:** {
**useDefault** - get flag from parent (default = False),
**True** - set as dynamic model,
**False** - set as a steady-state model.}""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Holdup construction flag",
            doc="""Indicates whether holdup terms should be constructed or not.
Must be True if dynamic = True,
**default** - False.
**Valid values:** {
**True** - construct holdup terms,
**False** - do not construct holdup terms}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )
    CONFIG.declare(
        "calculate_PA_SA_flows",
        ConfigValue(
            default=False,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether the primary air and secondary air calculations
have to be constructed or not,
**default** - False.
**Valid values:** {
**True** - calculate primary air and secondary air based on stoichiometric
ratio, PA to coal ratio, and lower stoichiometric ratio,
**False** - primary air and secondary air are inputs of the model.}""",
        ),
    )
    CONFIG.declare(
        "number_of_zones",
        ConfigValue(
            default=16,
            description="number of boiler zones",
            doc="number of boiler zones",
        ),
    )
    CONFIG.declare(
        "has_platen_superheater",
        ConfigValue(
            default=True,
            domain=Bool,
            description="True if boiler includes a platen superheater",
            doc="""Indicates if boiler includes a platen superheater,
**default** - True.
**Valid values:** {
**True** - include heat duty to platen superheater,
**False** - exclude heat duty to platen superheater.}""",
        ),
    )
    CONFIG.declare(
        "has_roof_superheater",
        ConfigValue(
            default=True,
            domain=Bool,
            description="True if roof is treated as a superheater",
            doc="""Indicates if roof is a section of superheater,
**default** - True.
**Valid values:** {
**True** - include heat duty to roof superheater,
**False** - exclude heat duty to roof superheater.}""",
        ),
    )
    CONFIG.declare(
        "surrogate_dictionary",
        ConfigValue(
            default=None,
            description="surrogate model dictionary",
            doc="""user must provide surrogate models or values for heat duty,
**default** - False.
**Valid values:** """,
        ),
    )

    def build(self):
        super(BoilerFiresideData, self).build()
        # Insert user provided custom surrogate model function
        # self.config.surrogate_function(self)
        if self.config.surrogate_dictionary is None:
            raise ConfigurationError(
                "{} - User needs to provide a dictionary of surrogates".format(
                    self.name
                )
            )

        self.primary_air = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )
        self.primary_air_moist = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )
        self.secondary_air = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )
        self.flue_gas = self.config.property_package.build_state_block(
            self.flowsheet().time, **self.config.property_package_args
        )

        self.add_port("primary_air_inlet", self.primary_air)
        self.add_port("secondary_air_inlet", self.secondary_air)
        self.add_port("flue_gas_outlet", self.flue_gas)
        # Construct performance equations
        self._make_params()
        self._make_vars()
        self._make_mass_balance()
        self._make_energy_balance()
        self._make_momentum_balance()
        self._import_surrogate_models()

    def _import_surrogate_models(self):

        data_dict = self.config.surrogate_dictionary
        if len(data_dict) != (
            len(self.zones)
            + self.config.has_platen_superheater
            + self.config.has_roof_superheater
            + 2  # flyash surrogate and NOx surrogate
        ):
            raise ConfigurationError(
                "User needs to provide the same number "
                "of surrogate models and water wall zones"
                "and/or platen sh and/or boiler roof"
            )

        # Surrogate model predictions
        # Constraints for heat duty in boiler water wall zones
        @self.Constraint(
            self.flowsheet().time,
            self.zones,
            doc="Surrogate model for heat loss" " to water wall zones",
        )
        def eq_surr_waterwall_heat(b, t, z):
            # Evaluating surrogate expression
            # pylint: disable=W0123
            return b.waterwall_heat[t, z] * b.fcorrection_heat_ww[t] == eval(
                data_dict[z]
            )

        if self.config.has_platen_superheater is True:

            @self.Constraint(
                self.flowsheet().time,
                doc="Surrogate model for heat loss" " to platen superheater",
            )
            def eq_surr_platen_heat(b, t):
                # Evaluating surrogate expression
                # pylint: disable=W0123
                return b.platen_heat[t] * b.fcorrection_heat_platen[t] == eval(
                    data_dict["pl"]
                )

        if self.config.has_roof_superheater is True:

            @self.Constraint(
                self.flowsheet().time,
                doc="Surrogate model for heat loss in " " the roof and backpass heater",
            )
            def eq_surr_roof_heat(b, t):
                # Evaluating surrogate expression
                # pylint: disable=W0123
                return b.roof_heat[t] * b.fcorrection_heat_ww[t] == eval(
                    data_dict["roof"]
                )

        # Constraints for unburned carbon
        @self.Constraint(
            self.flowsheet().time,
            doc="Surrogate model for" " mass fraction of unburned carbon",
        )
        def eq_surr_ln_ubc(b, t):
            # Evaluating surrogate expression
            # pylint: disable=W0123
            return b.ubc_in_flyash[t] == eval(data_dict["flyash"])

        # Constraints for NOx in mol fraction, surrogate model in PPM,
        # converted to mass fraction
        @self.Constraint(
            self.flowsheet().time,
            doc="NOx in mol fraction" "surrogate model must be in PPM",
        )
        def eq_surr_nox(b, t):
            # Evaluating surrogate expression
            # pylint: disable=W0123
            return b.frac_mol_NOx_fluegas[t] * 1e6 == eval(data_dict["NOx"])

        #                    # 1e6 conversion factor from PPM to mol fract

    def _make_params(self):
        """This section is for parameters within this model."""
        # atomic mass of elements involved in kg/mol
        self.atomic_mass_C = Param(initialize=0.0120107, doc="Atomic mass of C")
        self.atomic_mass_H = Param(initialize=0.00100795, doc="Atomic mass of H")
        self.atomic_mass_O = Param(initialize=0.0159994, doc="Atomic mass of O")
        self.atomic_mass_N = Param(initialize=0.0140067, doc="Atomic mass of N")
        self.atomic_mass_S = Param(initialize=0.0320652, doc="Atomic mass of S")
        # mole fractions of O2, H2O, CO2, SO2, and NO in air
        # at 25 C and relative humidity of 25% with SO2 and NO at 0.1 ppm
        # change air compostion accordingly if relative humidity changes
        # N2 mole fraction not needed since summation of mole fractions of
        # all six species is 1. It is 0.7839958
        if self.config.calculate_PA_SA_flows is True:
            self.mole_frac_air = Param(
                self.config.property_package.component_list,
                mutable=True,
                initialize={
                    "O2": 0.20784,
                    "N2": 0.783994,
                    "NO": 0.0000001,
                    "CO2": 0.0003373,
                    "H2O": 0.0078267,
                    "SO2": 0.0000001,
                },
                doc="Mole fraction of air species",
            )

    def _make_vars(self):
        """This section is for variables within this model."""
        # Number of waterwall zones is given by config,
        # Optionally the model could contain a platen and a roof superheater

        self.deltaP = Var(
            self.flowsheet().time,
            initialize=1000,
            doc="Pressure drop of secondary air " "through windbox and burner [Pa]",
        )

        self.zones = RangeSet(self.config.number_of_zones)

        self.fcorrection_heat_ww = Var(
            self.flowsheet().time,
            initialize=1,
            doc="Correction factor " "for waterwall heat duty",
        )

        if self.config.has_platen_superheater is True:
            self.fcorrection_heat_platen = Var(
                self.flowsheet().time,
                initialize=1,
                doc="Correction factor for " "platen SH heat duty",
            )

        # wall temperatures of water wall zones
        self.wall_temperature_waterwall = Var(
            self.flowsheet().time,
            self.zones,
            initialize=700.0,
            doc="Wall temperature [K] in " "Waterwall zones",
        )

        # heat duties for water wall zones
        self.waterwall_heat = Var(
            self.flowsheet().time,
            self.zones,
            initialize=2.0e7,
            doc="Heat duty [W] or heat loss " "to waterwall zones",
        )

        if self.config.has_platen_superheater is True:
            # heat duty to platen super heater
            self.platen_heat = Var(
                self.flowsheet().time,
                initialize=6.0e7,
                doc="Platen superheater heat duty [W]",
            )
            # wall temperature of platen superheater
            self.wall_temperature_platen = Var(
                self.flowsheet().time,
                initialize=800.0,
                doc="Platen superheater" " wall temperature [K]",
            )

        if self.config.has_roof_superheater is True:
            # heat duty to roof superheater
            self.roof_heat = Var(
                self.flowsheet().time,
                initialize=5e6,
                doc="Roof superheater heat duty [W]",
            )
            # wall temperature of roof superheater
            self.wall_temperature_roof = Var(
                self.flowsheet().time,
                initialize=800.0,
                doc="Roof superheater " "wall temperature [K]",
            )

        # PA/coal temperature, usually fixed around 150 F
        self.temperature_coal = Var(
            self.flowsheet().time, initialize=338.7, doc="Coal temperature in K"
        )

        # overall Stoichiometric ratio, used to calculate total
        # combustion air flowrate
        # If SR is a function of load or raw coal flow rate,
        # specify constraint in flowsheet model
        self.SR = Var(
            self.flowsheet().time,
            initialize=1.15,
            doc="Overall furnace Stoichiometric ratio - SR",
        )

        # lower furnace Stoichiometric ratio
        self.SR_lf = Var(
            self.flowsheet().time,
            initialize=1.15,
            doc="Lower furnace Stoichiometric ratio" " - SR excluding overfire air",
        )

        # PA to coal mass flow ratio,
        # typically 2.0 depending on load or mill curve
        # usually this is an input of surrogate model
        self.ratio_PA2coal = Var(
            self.flowsheet().time, initialize=2.0, doc="Primary air to coal ratio"
        )

        # mass flow rate of raw coal fed to mill
        # (raw coal flow without moisture vaporization in mill)
        self.flowrate_coal_raw = Var(
            self.flowsheet().time, initialize=25.0, doc="Raw coal mass flowrate [kg/s]"
        )

        # mass flow rate of coal to burners after moisture vaporization in mill
        self.flowrate_coal_burner = Var(
            self.flowsheet().time,
            initialize=20.0,
            doc="Mass flowrate coal to burners "
            "with moisture"
            " partially vaporized in mill [kg/s]",
        )

        # raw coal moisture mass fraction (on as received basis),
        # can change with time to represent HHV change on as received basis
        self.mf_H2O_coal_raw = Var(
            self.flowsheet().time,
            initialize=0.15,
            doc="Raw coal mass fraction of" " moisture on as received basis",
        )

        # moisture mass fraction of coal to burners after mill,
        # calculated based on fraction of moistures vaporized in mill
        self.mf_H2O_coal_burner = Var(
            self.flowsheet().time,
            initialize=0.15,
            doc="Mass fraction of moisture" " on as received basis",
        )

        # Fraction of moisture vaporized in mill,
        # set in flowsheet as a function of coal flow rate, default is 0.6
        self.frac_moisture_vaporized = Var(
            self.flowsheet().time,
            initialize=0.6,
            doc="Fraction of coal" " moisture vaporized in mill",
        )

        # Vaporized moisture mass flow rate
        @self.Expression(
            self.flowsheet().time, doc="Vaporized moisture mass flowrate [kg/s]"
        )
        def flowrate_moist_vaporized(b, t):
            return (
                b.flowrate_coal_raw[t]
                * b.mf_H2O_coal_raw[t]
                * b.frac_moisture_vaporized[t]
            )

        # Elemental composition of dry coal, assuming fixed over time
        # They are usually fixed as user inputs
        self.mf_C_coal_dry = Var(initialize=0.5, doc="Mass fraction of C on dry basis")

        self.mf_H_coal_dry = Var(initialize=0.05, doc="Mass fraction of H on dry basis")

        self.mf_O_coal_dry = Var(initialize=0.1, doc="Mass fraction of O on dry basis")

        self.mf_N_coal_dry = Var(initialize=0.1, doc="Mass fraction of N on dry basis")

        self.mf_S_coal_dry = Var(initialize=0.05, doc="Mass fraction of S on dry basis")

        self.mf_Ash_coal_dry = Var(
            initialize=0.2, doc="Mass fraction of Ash on dry basis"
        )

        # High heating value of dry coal, usually as a fixed user input
        self.hhv_coal_dry = Var(
            initialize=1e7, doc="High heating value (HHV)" "of coal on dry basis J/kg"
        )

        # Fraction of unburned carbon (actually all organic elements) in
        # flyash, predicted by surrogate model.
        # When generating the surrogate, sum up all elements in
        # flyash from fireside boiler model outputs
        self.ubc_in_flyash = Var(
            self.flowsheet().time,
            initialize=0.01,
            doc="Unburned carbon and" " other organic elements in fly ash",
        )

        # mole fraction of NO in flue gas, predicted by surrogate model
        # usually mole fraction is very close to mass fraction of NO
        self.frac_mol_NOx_fluegas = Var(
            self.flowsheet().time,
            initialize=1e-4,
            doc="Mole fraction of NOx in flue gas",
        )

        # NOx in lb/MMBTU, NO is converted to NO2 as NOx
        @self.Expression(self.flowsheet().time, doc="NOx in lb/MMBTU")
        def nox_lb_mmbtu(b, t):
            return (
                b.flue_gas_outlet.flow_mol_comp[t, "NO"]
                * 0.046
                * 2.20462
                / (
                    b.flowrate_coal_raw[t]
                    * (1 - b.mf_H2O_coal_raw[t])
                    * b.hhv_coal_dry
                    / 1.054e9
                )
            )

        # coal flow rate after mill
        @self.Constraint(
            self.flowsheet().time, doc="Coal flow rate to burners after mill"
        )
        def flowrate_coal_burner_eqn(b, t):
            return (
                b.flowrate_coal_burner[t]
                == b.flowrate_coal_raw[t] - b.flowrate_moist_vaporized[t]
            )

        # moisture mass fraction in coal after mill
        @self.Constraint(
            self.flowsheet().time, doc="Moisture mass fraction for coal after mill"
        )
        def mf_H2O_coal_burner_eqn(b, t):
            return (
                b.flowrate_coal_raw[t] * b.mf_H2O_coal_raw[t]
                == b.flowrate_coal_burner[t] * b.mf_H2O_coal_burner[t]
                + b.flowrate_moist_vaporized[t]
            )

        # fraction of daf elements on dry basis
        @self.Expression(doc="Mass fraction of dry ash free (daf)")
        def mf_daf_dry(b):
            return 1 - b.mf_Ash_coal_dry

        @self.Expression(self.flowsheet().time, doc="Ash mass flow rate kg/s")
        def flowrate_ash(b, t):
            return (
                b.flowrate_coal_raw[t] * (1 - b.mf_H2O_coal_raw[t]) * b.mf_Ash_coal_dry
            )

        @self.Expression(
            self.flowsheet().time,
            doc="Dry ash free - daf_coal flow rate " "in fuel fed to the boiler kg/s",
        )
        def flowrate_daf_fuel(b, t):
            return b.flowrate_coal_raw[t] * (1 - b.mf_H2O_coal_raw[t]) * b.mf_daf_dry

        @self.Expression(
            self.flowsheet().time, doc="Dry ash free (daf) coal flow rate in fly ash"
        )
        def flowrate_daf_flyash(b, t):
            return b.flowrate_ash[t] / (1.0 - b.ubc_in_flyash[t]) * b.ubc_in_flyash[t]

        @self.Expression(self.flowsheet().time, doc="Burned daf coal flow rate")
        def flowrate_daf_burned(b, t):
            return b.flowrate_daf_fuel[t] - b.flowrate_daf_flyash[t]

        @self.Expression(self.flowsheet().time, doc="Coal burnout")
        def coal_burnout(b, t):
            return b.flowrate_daf_burned[t] / b.flowrate_daf_fuel[t]

        @self.Expression(doc="Dry ash free - daf C mass fraction")
        def mf_C_daf(b):
            return b.mf_C_coal_dry / b.mf_daf_dry

        @self.Expression(doc="Dry ash free - daf H mass fraction")
        def mf_H_daf(b):
            return b.mf_H_coal_dry / b.mf_daf_dry

        @self.Expression(doc="Dry ash free - daf O mass fraction")
        def mf_O_daf(b):
            return b.mf_O_coal_dry / b.mf_daf_dry

        @self.Expression(doc="Dry ash free - daf N mass fraction")
        def mf_N_daf(b):
            return b.mf_N_coal_dry / b.mf_daf_dry

        @self.Expression(doc="Dry ash free - daf S mass fraction")
        def mf_S_daf(b):
            return b.mf_S_coal_dry / b.mf_daf_dry

        @self.Expression(doc="Dry ash free high heating value in J/kg")
        def hhv_daf(b):
            return b.hhv_coal_dry / b.mf_daf_dry

        @self.Expression(
            doc="Heat of combustion at constant " "pressure on daf basis in J/kg"
        )
        # Note that measure HHV is the heat of combustion at constant volume
        def dhcoal(b):
            return -b.hhv_daf + const.gas_constant * 298.15 / 2 * (
                -b.mf_H_daf / 2 / b.atomic_mass_H
                + b.mf_O_daf / b.atomic_mass_O
                + b.mf_N_daf / b.atomic_mass_N
            )

        @self.Expression(doc="Carbon heat of reaction - dhc J/kg")
        def dhc(b):
            return -94052 * 4.184 / b.atomic_mass_C * b.mf_C_daf

        @self.Expression(doc="Hydrogen heat of reaction - dhh in J/kg")
        def dhh(b):
            return -68317.4 * 4.184 / b.atomic_mass_H / 2 * b.mf_H_daf

        @self.Expression(doc="Sulfur heat of reaction - dhs in J/kg")
        def dhs(b):
            return -70940 * 4.184 / b.atomic_mass_S * b.mf_S_daf

        @self.Expression(doc="Heat of formation for daf coal in J/kg")
        def hf_daf(b):
            return b.dhc + b.dhh + b.dhs - b.dhcoal

        @self.Expression(
            self.flowsheet().time,
            doc="Heat of formation of " "moisture-containing coal to burners",
        )
        def hf_coal(b, t):
            return b.hf_daf * (
                1 - b.mf_H2O_coal_burner[t]
            ) * b.mf_daf_dry + b.mf_H2O_coal_burner[t] * (-68317.4) * 4.184 / (
                b.atomic_mass_H * 2 + b.atomic_mass_O
            )

        @self.Expression(doc="Average atomic mass of daf coal")
        def am_daf(b):
            return 1 / (
                b.mf_C_daf / b.atomic_mass_C
                + b.mf_H_daf / b.atomic_mass_H
                + b.mf_O_daf / b.atomic_mass_O
                + b.mf_N_daf / b.atomic_mass_N
                + b.mf_S_daf / b.atomic_mass_S
            )

        @self.Expression(
            self.flowsheet().time,
            doc="Gt1 or Einstein quantum theory function for daf "
            "coal sensible heat calculation",
        )
        def gt1(b, t):
            return 1 / (exp(380 / b.temperature_coal[t]) - 1)

        # since flyash and flue gas outlet temperature is not fixed
        self.gt1_flyash = Var(
            self.flowsheet().time,
            initialize=1,
            doc="Gt1 or Einstein quantum theory function" " for daf part of flyash",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Gt1 or Einstein quantum theory " "for daf part of flyash",
        )
        def gt1_flyash_eqn(b, t):
            return b.gt1_flyash[t] * (exp(380 / b.flue_gas[t].temperature) - 1) == 1

        @self.Expression(
            self.flowsheet().time,
            doc="Gt2 or Einstein quantum theory "
            "function for daf coal "
            "sensible heat calculation for high temperature",
        )
        def gt2(b, t):
            return 1 / (exp(1800 / b.temperature_coal[t]) - 1)

        # declare variable rather than use expression
        # since flyash and flue gas outlet temperature is not fixed
        self.gt2_flyash = Var(
            self.flowsheet().time,
            initialize=1,
            doc="Gt2 or Einstein quantum theory function " "for daf part of flyash",
        )

        @self.Constraint(
            self.flowsheet().time,
            doc="Gt2 or Einstein quantum theory function " "for daf part of flyash",
        )
        def gt2_flyash_eqn(b, t):
            return b.gt2_flyash[t] * (exp(1800 / b.flue_gas[t].temperature) - 1) == 1

        @self.Expression(self.flowsheet().time, doc="Sensible heat of daf coal")
        def hs_daf(b, t):
            return (
                const.gas_constant
                / b.am_daf
                * (380 * (b.gt1[t] - 0.3880471566) + 3600 * (b.gt2[t] - 0.002393883))
            )

        @self.Expression(
            self.flowsheet().time, doc="Sensible heat for daf part of flyash"
        )
        def hs_daf_flyash(b, t):
            return (
                const.gas_constant
                / b.am_daf
                * (
                    380 * (b.gt1_flyash[t] - 0.3880471566)
                    + 3600 * (b.gt2_flyash[t] - 0.002393883)
                )
            )

        @self.Expression(self.flowsheet().time, doc="Sensible heat of coal to burners")
        def hs_coal(b, t):
            return (
                (1 - b.mf_H2O_coal_burner[t]) * b.mf_daf_dry * b.hs_daf[t]
                + b.mf_H2O_coal_burner[t] * 4184 * (b.temperature_coal[t] - 298.15)
                + (1 - b.mf_H2O_coal_burner[t])
                * b.mf_Ash_coal_dry
                * (
                    593 * (b.temperature_coal[t] - 298.15)
                    + 0.293 * (b.temperature_coal[t] ** 2 - 298.15**2)
                )
            )

        @self.Expression(
            self.flowsheet().time,
            doc="Total enthalpy of " "moisture-containing coal to burners",
        )
        def h_coal(b, t):
            return b.hs_coal[t] + b.hf_coal[t]

        # variable for O2 mole fraction in flue gas on dry basis
        # it can be used for data reconciliation and parameter estimation
        self.fluegas_o2_pct_dry = Var(
            self.flowsheet().time, initialize=3, doc="Mol percent of O2 on dry basis"
        )

    def _make_mass_balance(self):
        # PA flow rate calculated based on PA/coal ratio
        @self.Constraint(
            self.flowsheet().time, doc="Ratio of PA mass flow to raw coal mass flow"
        )
        def ratio_PA2coal_eqn(b, t):
            return (
                b.primary_air[t].flow_mass
                == b.flowrate_coal_raw[t] * b.ratio_PA2coal[t]
            )

        @self.Expression(self.flowsheet().time, doc="C molar flow from coal")
        def molflow_C_fuel(b, t):
            return b.flowrate_daf_fuel[t] * b.mf_C_daf / b.atomic_mass_C

        @self.Expression(self.flowsheet().time, doc="H molar flow from coal")
        def molflow_H_fuel(b, t):
            return b.flowrate_daf_fuel[t] * b.mf_H_daf / b.atomic_mass_H

        @self.Expression(self.flowsheet().time, doc="O molar flow from coal")
        def molflow_O_fuel(b, t):
            return b.flowrate_daf_fuel[t] * b.mf_O_daf / b.atomic_mass_O

        @self.Expression(self.flowsheet().time, doc="N molar flow from coal")
        def molflow_N_fuel(b, t):
            return b.flowrate_daf_fuel[t] * b.mf_N_daf / b.atomic_mass_N

        @self.Expression(self.flowsheet().time, doc="S molar flow from coal")
        def molflow_S_fuel(b, t):
            return b.flowrate_daf_fuel[t] * b.mf_S_daf / b.atomic_mass_S

        if self.config.calculate_PA_SA_flows is True:
            # Calculate PA component molar flow based on air mol fractions
            # specified as model parameter
            # Overwrite the composition from inlets
            # Let N2 molar flow calculated by balance
            @self.Constraint(
                self.flowsheet().time,
                self.config.property_package.component_list,
                doc="PA component molar flow",
            )
            def molar_flow_PA_eqn(b, t, j):
                if j == "N2":
                    return Constraint.Skip
                else:
                    return (
                        b.primary_air[t].flow_mol_comp[j]
                        == b.primary_air[t].flow_mol * b.mole_frac_air[j]
                    )

            # Calculate SA component molar flow based on air mol fractions
            # specified as model parameter
            # Overwrite the composition from inlets
            # Let N2 molar flow calculated by balance
            @self.Constraint(
                self.flowsheet().time,
                self.config.property_package.component_list,
                doc="SA component molar flow",
            )
            def molar_flow_SA_eqn(b, t, j):
                if j == "N2":
                    return Constraint.Skip
                else:
                    return (
                        b.secondary_air[t].flow_mol_comp[j]
                        == b.secondary_air[t].flow_mol * b.mole_frac_air[j]
                    )

        # calculate molar flow of primary_air_moist stream
        @self.Constraint(
            self.flowsheet().time, self.config.property_package.component_list
        )
        def primary_air_moist_comp_flow_eqn(b, t, j):
            if j == "H2O":
                return (
                    b.primary_air[t].flow_mol_comp["H2O"]
                    + b.flowrate_moist_vaporized[t]
                    / (b.atomic_mass_H * 2 + b.atomic_mass_O)
                ) == b.primary_air_moist[t].flow_mol_comp["H2O"]
            else:
                return (
                    b.primary_air[t].flow_mol_comp[j]
                    == b.primary_air_moist[t].flow_mol_comp[j]
                )

        # total combustion air mass flow rate
        @self.Expression(
            self.flowsheet().time, doc="Total combustion air mass flow rate"
        )
        def flow_mass_TCA(b, t):
            return b.primary_air[t].flow_mass + b.secondary_air[t].flow_mass

        # overall Stoichiometric ratio (SR)
        @self.Constraint(self.flowsheet().time, doc="SR equation")
        def SR_eqn(b, t):
            return (
                b.SR[t]
                * (
                    b.molflow_C_fuel[t]
                    + b.molflow_H_fuel[t] / 4
                    + b.molflow_S_fuel[t]
                    - b.molflow_O_fuel[t] / 2
                )
                == b.primary_air[t].flow_mol_comp["O2"]
                + b.secondary_air[t].flow_mol_comp["O2"]
            )

        @self.Expression(self.flowsheet().time, doc="C mole flow")
        def molflow_C_fluegas(b, t):
            return (
                b.flowrate_daf_burned[t] * b.mf_C_daf / b.atomic_mass_C
                + b.primary_air_moist[t].flow_mol_comp["CO2"]
                + b.secondary_air[t].flow_mol_comp["CO2"]
            )

        @self.Expression(self.flowsheet().time, doc="H mole flow")
        def molflow_H_fluegas(b, t):
            return (
                b.flowrate_daf_burned[t] * b.mf_H_daf / b.atomic_mass_H
                + b.primary_air_moist[t].flow_mol_comp["H2O"] * 2
                + b.secondary_air[t].flow_mol_comp["H2O"] * 2
                + b.flowrate_coal_burner[t]
                * b.mf_H2O_coal_burner[t]
                / (b.atomic_mass_H * 2 + b.atomic_mass_O)
                * 2
            )

        @self.Expression(self.flowsheet().time, doc="O mole flow")
        def molflow_O_fluegas(b, t):
            return (
                b.flowrate_daf_burned[t] * b.mf_O_daf / b.atomic_mass_O
                + b.primary_air_moist[t].flow_mol_comp["O2"] * 2
                + b.primary_air_moist[t].flow_mol_comp["CO2"] * 2
                + b.primary_air_moist[t].flow_mol_comp["H2O"]
                + b.primary_air_moist[t].flow_mol_comp["SO2"] * 2
                + b.primary_air_moist[t].flow_mol_comp["NO"]
                + b.secondary_air[t].flow_mol_comp["O2"] * 2
                + b.secondary_air[t].flow_mol_comp["CO2"] * 2
                + b.secondary_air[t].flow_mol_comp["H2O"]
                + b.secondary_air[t].flow_mol_comp["SO2"] * 2
                + b.secondary_air[t].flow_mol_comp["NO"]
                + b.flowrate_coal_burner[t]
                * b.mf_H2O_coal_burner[t]
                / (b.atomic_mass_H * 2 + b.atomic_mass_O)
            )

        @self.Expression(self.flowsheet().time, doc="N mole flow")
        def molflow_N_fluegas(b, t):
            return (
                b.flowrate_daf_burned[t] * b.mf_N_daf / b.atomic_mass_N
                + b.primary_air_moist[t].flow_mol_comp["N2"] * 2
                + b.primary_air_moist[t].flow_mol_comp["NO"]
                + b.secondary_air[t].flow_mol_comp["N2"] * 2
                + b.secondary_air[t].flow_mol_comp["NO"]
            )

        @self.Expression(self.flowsheet().time, doc="S mole flow")
        def molflow_S_fluegas(b, t):
            return (
                b.flowrate_daf_burned[t] * b.mf_S_daf / b.atomic_mass_S
                + b.primary_air_moist[t].flow_mol_comp["SO2"]
                + b.secondary_air[t].flow_mol_comp["SO2"]
            )

        # calculate flue gas flow component at flue_gas_outlet
        # NO mole fraction in flue gas is given by NOx surrogate model
        @self.Constraint(self.flowsheet().time, doc="NO at flue gas outlet")
        def NO_eqn(b, t):
            return (
                b.flue_gas[t].flow_mol_comp["NO"]
                == b.frac_mol_NOx_fluegas[t] * b.flue_gas[t].flow_mol
            )

        # N2 at outlet
        @self.Constraint(self.flowsheet().time, doc="N2 at outlet")
        def N2_eqn(b, t):
            return (
                b.flue_gas_outlet.flow_mol_comp[t, "N2"]
                == +b.molflow_N_fluegas[t] / 2
                - b.flue_gas_outlet.flow_mol_comp[t, "NO"] / 2
            )

        # SO2 at outlet
        @self.Constraint(self.flowsheet().time, doc="SO2 at outlet")
        def SO2_eqn(b, t):
            return b.flue_gas_outlet.flow_mol_comp[t, "SO2"] == b.molflow_S_fluegas[t]

        # H2O at outlet
        @self.Constraint(self.flowsheet().time, doc="H2O at outlet")
        def H2O_eqn(b, t):
            return (
                b.flue_gas_outlet.flow_mol_comp[t, "H2O"] == b.molflow_H_fluegas[t] / 2
            )

        # CO2 at outlet
        @self.Constraint(self.flowsheet().time, doc="CO2 at outlet")
        def CO2_eqn(b, t):
            return b.flue_gas_outlet.flow_mol_comp[t, "CO2"] == b.molflow_C_fluegas[t]

        # O2 at outlet
        @self.Constraint(self.flowsheet().time, doc="O2 at outlet")
        def O2_eqn(b, t):
            return (
                b.flue_gas_outlet.flow_mol_comp[t, "O2"]
                == (
                    b.molflow_O_fluegas[t]
                    - b.molflow_C_fluegas[t] * 2
                    - b.molflow_H_fluegas[t] / 2
                    - b.molflow_S_fluegas[t] * 2
                    - b.flue_gas_outlet.flow_mol_comp[t, "NO"]
                )
                / 2
            )

        # constraint for mol percent of O2 in flue gas on dry basis
        @self.Constraint(
            self.flowsheet().time, doc="Mol percent of O2 in flue gas on dry basis"
        )
        def fluegas_o2_pct_dry_eqn(b, t):
            return (
                b.fluegas_o2_pct_dry[t]
                * (
                    b.flue_gas_outlet.flow_mol_comp[t, "O2"]
                    + b.flue_gas_outlet.flow_mol_comp[t, "N2"]
                    + b.flue_gas_outlet.flow_mol_comp[t, "CO2"]
                    + b.flue_gas_outlet.flow_mol_comp[t, "SO2"]
                    + b.flue_gas_outlet.flow_mol_comp[t, "NO"]
                )
                / 100
                == b.flue_gas_outlet.flow_mol_comp[t, "O2"]
            )

        # mass flow rate of flyash containing unburned fuel
        @self.Expression(self.flowsheet().time, doc="Flyash mass flow rate  kg/s")
        def flow_mass_flyash(b, t):
            return b.flowrate_daf_flyash[t] + b.flowrate_ash[t]

    def _make_momentum_balance(self):
        # flue gas pressure is secondary air pressure
        # - pressure drop through windbox
        # and burner secondary air register
        @self.Constraint(self.flowsheet().time, doc="Flue gas pressure in Pascals")
        def flue_gas_pressure_eqn(b, t):
            return (
                b.flue_gas[t].pressure * 1e-5
                == (b.secondary_air[t].pressure - b.deltaP[t]) * 1e-5
            )

        # set pressure of primary air with moisture at mill outlet
        @self.Constraint(self.flowsheet().time, doc="Mill outlet pressure")
        def primary_air_moist_pressure_eqn(b, t):
            return (
                b.primary_air[t].pressure * 1e-5
                == b.primary_air_moist[t].pressure * 1e-5
            )

    def _make_energy_balance(self):
        # temperature of primary_air_moist is equal to coal temperature
        # leaving mill, ignore the temperature of primary_air_inlet in energy
        # balance equation
        @self.Constraint(
            self.flowsheet().time, doc="Temperature of primary air leaving mill"
        )
        def primary_air_moist_temperature_eqn(b, t):
            return b.primary_air_moist[t].temperature == b.temperature_coal[t]

        # overall energy balance to calculate FEGT
        @self.Constraint(self.flowsheet().time, doc="Temperature at flue gas outlet")
        def flue_gas_temp_eqn(b, t):
            return (
                b.primary_air_moist[t].flow_mol * b.primary_air_moist[t].enth_mol
                + b.secondary_air[t].flow_mol * b.secondary_air[t].enth_mol
                + b.h_coal[t] * b.flowrate_coal_burner[t]
            ) == (
                sum(b.waterwall_heat[t, j] for j in b.zones)
                + (
                    b.platen_heat[t]
                    if self.config.has_platen_superheater is True
                    else 0
                )
                + (b.roof_heat[t] if self.config.has_roof_superheater is True else 0)
                + b.flue_gas[t].flow_mol * b.flue_gas[t].enth_mol
                # sensible heat of ash
                + b.flowrate_ash[t]
                * (
                    593 * (b.flue_gas[t].temperature - 298.15)
                    + 0.293 * (b.flue_gas[t].temperature ** 2.0 - 298.15**2.0)
                )
                # sensible heat of unburned fuel
                + b.flowrate_daf_flyash[t] * b.hs_daf_flyash[t]
            )

        # expression to calculate total heat duty of waterwall
        @self.Expression(
            self.flowsheet().time, doc="Total heat duty of all waterwall zones"
        )
        def heat_total_ww(b, t):
            return sum(b.waterwall_heat[t, j] for j in b.zones)

        # expression to calculate total heat loss through
        # waterwall, platen SH, and roof SH
        @self.Expression(self.flowsheet().time, doc="Total heat duty")
        def heat_total(b, t):
            return (
                b.heat_total_ww[t]
                + (
                    b.platen_heat[t]
                    if self.config.has_platen_superheater is True
                    else 0
                )
                + (b.roof_heat[t] if self.config.has_roof_superheater is True else 0)
            )

    def initialize_build(
        blk,
        state_args_PA=None,
        state_args_SA=None,
        outlvl=idaeslog.NOTSET,
        solver=None,
        optarg=None,
    ):
        """
        Initialization routine.
        1.- initialize state blocks, using an initial guess for inlet
        primary air and secondary air.
        2.- Use PA and SA values to guess flue gas component molar flowrates,
        Temperature, and Pressure. Initialize flue gas state block.
        3.- Then, solve complete model.

        Keyword Arguments:
            state_args_PA : a dict of arguments to be passed to the property
                           package(s) for the primary air state block to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            state_args_SA : a dict of arguments to be passed to the property
                           package(s) for secondary air state block to
                           provide an initial state for initialization
                           (see documentation of the specific property package)
                           (default = None).
            outlvl : sets output level of initialisation routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None, use default solver)

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(blk.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(blk.name, outlvl, tag="unit")

        # Create solver
        opt = get_solver(solver, optarg)

        # ---------------------------------------------------------------------
        # Initialize inlet property blocks
        blk.primary_air.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_PA
        )
        blk.primary_air_moist.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_PA
        )
        blk.secondary_air.initialize(
            outlvl=outlvl, optarg=optarg, solver=solver, state_args=state_args_SA
        )
        init_log.info_high("Initialization Step 1 Complete.")

        state_args = {
            "flow_mol_comp": {
                "H2O": (
                    blk.primary_air_inlet.flow_mol_comp[0, "H2O"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "H2O"].value
                ),
                "CO2": (
                    blk.primary_air_inlet.flow_mol_comp[0, "CO2"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "CO2"].value
                ),
                "N2": (
                    blk.primary_air_inlet.flow_mol_comp[0, "N2"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "N2"].value
                ),
                "O2": (
                    blk.primary_air_inlet.flow_mol_comp[0, "O2"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "O2"].value
                ),
                "SO2": (
                    blk.primary_air_inlet.flow_mol_comp[0, "SO2"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "SO2"].value
                ),
                "NO": (
                    blk.primary_air_inlet.flow_mol_comp[0, "NO"].value
                    + blk.secondary_air_inlet.flow_mol_comp[0, "NO"].value
                ),
            },
            "temperature": 1350.00,
            "pressure": blk.primary_air_inlet.pressure[0].value,
        }
        # initialize flue gas outlet
        blk.flue_gas.initialize(state_args=state_args, outlvl=outlvl, solver=solver)
        init_log.info_high("Initialization Step 2 Complete.")

        if blk.config.calculate_PA_SA_flows is False:
            # Option 1: given PA and SA component flow rates - fixed inlets
            # fix inlet component molar flow rates
            # unfix ratio_PA2coal, SR, and fluegas_o2_pct_dry
            blk.primary_air_inlet.flow_mol_comp[...].fix()
            blk.secondary_air_inlet.flow_mol_comp[...].fix()
            blk.ratio_PA2coal.unfix()
            blk.SR.unfix()
            blk.fluegas_o2_pct_dry.unfix()
            dof = degrees_of_freedom(blk)

        else:
            # Option 2: SR, ratioPA2_coal to estimate TCA, PA, SA
            # unfix component molar flow rates, but keep T and P fixed.
            blk.primary_air_inlet.flow_mol_comp[:, :].unfix()
            blk.secondary_air_inlet.flow_mol_comp[:, :].unfix()
            dof = degrees_of_freedom(blk)

        if not dof == 0:
            raise ConfigurationError("User needs to check degrees of freedom")

        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(blk, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        init_log.info("Initialization Complete.")

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        # set a default waterwall zone heat scaling factor
        for v in self.waterwall_heat.values():
            if iscale.get_scaling_factor(v, warning=True) is None:
                iscale.set_scaling_factor(v, 1e-7)

        # set a default platen heat scaling factor
        if self.config.has_platen_superheater is True:
            for v in self.platen_heat.values():
                if iscale.get_scaling_factor(v, warning=True) is None:
                    iscale.set_scaling_factor(v, 1e-7)

        # set a default roof heat scaling factor
        if self.config.has_roof_superheater is True:
            for v in self.roof_heat.values():
                if iscale.get_scaling_factor(v, warning=True) is None:
                    iscale.set_scaling_factor(v, 1e-6)

        # set waterwall heat constraint scaling factor
        for t, c in self.eq_surr_waterwall_heat.items():
            sf = iscale.get_scaling_factor(
                self.waterwall_heat[t], default=1e-7, warning=True
            )
            iscale.constraint_scaling_transform(c, sf, overwrite=False)

        # set platen heat constraint scaling factor
        if self.config.has_platen_superheater is True:
            for t, c in self.eq_surr_platen_heat.items():
                sf = iscale.get_scaling_factor(
                    self.platen_heat[t], default=1e-7, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        # set roof heat constraint scaling factor
        if self.config.has_roof_superheater is True:
            for t, c in self.eq_surr_roof_heat.items():
                sf = iscale.get_scaling_factor(
                    self.roof_heat[t], default=1e-6, warning=True
                )
                iscale.constraint_scaling_transform(c, sf, overwrite=False)

        # set flue gas temperature constraint scaling factor
        for t, c in self.flue_gas_temp_eqn.items():
            sf = iscale.get_scaling_factor(
                self.platen_heat[t], default=1e-7, warning=True
            )
            iscale.constraint_scaling_transform(c, sf, overwrite=False)



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Standard IDAES flash model.
"""
# Import Python libraries
import logging
from pandas import DataFrame

# Import Pyomo libraries
from pyomo.environ import Constraint, Reference
from pyomo.common.config import ConfigBlock, ConfigValue, In, Bool
from pyomo.network import Port

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
)
from idaes.models.unit_models.separator import (
    Separator,
    SplittingType,
    EnergySplittingType,
)

from idaes.core.util.config import is_physical_parameter_block
from idaes.core.util.units_of_measurement import report_quantity


__author__ = "Andrew Lee, Jaffer Ghouse"


# Set up logger
logger = logging.getLogger("idaes.unit_model")


@declare_process_block_class("Flash")
class FlashData(UnitModelBlockData):
    """
    Standard Flash Unit Model Class
    """

    CONFIG = ConfigBlock()
    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
**default** = False. Flash units do not support dynamic behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
**default** - False. Flash units do not have defined volume, thus
this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
**default** - MaterialBalanceType.useDefault.
**Valid values:** {
**MaterialBalanceType.useDefault - refer to property package for default
balance type
**MaterialBalanceType.none** - exclude material balances,
**MaterialBalanceType.componentPhase** - use phase component balances,
**MaterialBalanceType.componentTotal** - use total component balances,
**MaterialBalanceType.elementTotal** - use total element balances,
**MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
**default** - EnergyBalanceType.useDefault.
**Valid values:** {
**EnergyBalanceType.useDefault - refer to property package for default
balance type
**EnergyBalanceType.none** - exclude energy balances,
**EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
**EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
**EnergyBalanceType.energyTotal** - single energy balance for material,
**EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
**default** - MomentumBalanceType.pressureTotal.
**Valid values:** {
**MomentumBalanceType.none** - exclude momentum balances,
**MomentumBalanceType.pressureTotal** - single pressure balance for material,
**MomentumBalanceType.pressurePhase** - pressure balances for each phase,
**MomentumBalanceType.momentumTotal** - single momentum balance for material,
**MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "energy_split_basis",
        ConfigValue(
            default=EnergySplittingType.equal_temperature,
            domain=EnergySplittingType,
            description="Type of constraint to write for energy splitting",
            doc="""Argument indicating basis to use for splitting energy this is
not used for when ideal_separation == True.
**default** - EnergySplittingType.equal_temperature.
**Valid values:** {
**EnergySplittingType.equal_temperature** - outlet temperatures equal inlet
**EnergySplittingType.equal_molar_enthalpy** - outlet molar enthalpies equal
inlet,
**EnergySplittingType.enthalpy_split** - apply split fractions to enthalpy
flows.}""",
        ),
    )
    CONFIG.declare(
        "ideal_separation",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Ideal splitting flag",
            doc="""Argument indicating whether ideal splitting should be used.
Ideal splitting assumes perfect separation of material, and attempts to
avoid duplication of StateBlocks by directly partitioning outlet flows to
ports,
**default** - True.
**Valid values:** {
**True** - use ideal splitting methods. Cannot be combined with
has_phase_equilibrium = True,
**False** - use explicit splitting equations with split fractions.}""",
        ),
    )
    CONFIG.declare(
        "has_heat_transfer",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Heat transfer term construction flag",
            doc="""Indicates whether terms for heat transfer should be constructed,
**default** - False.
**Valid values:** {
**True** - include heat transfer terms,
**False** - exclude heat transfer terms.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=True,
            domain=Bool,
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
constructed,
**default** - True.
**Valid values:** {
**True** - include pressure change terms,
**False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
**default** - useDefault.
**Valid values:** {
**useDefault** - use default package from parent model or flowsheet,
**PropertyParameterObject** - a PropertyParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
and used when constructing these,
**default** - None.
**Valid values:** {
see property package for documentation.}""",
        ),
    )

    def build(self):
        """
        Begin building model (pre-DAE transformation).

        Args:
            None

        Returns:
            None
        """
        # Call UnitModel.build to setup dynamics
        super(FlashData, self).build()

        # Build Control Volume
        self.control_volume = ControlVolume0DBlock(
            dynamic=self.config.dynamic,
            has_holdup=self.config.has_holdup,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )

        self.control_volume.add_state_blocks(has_phase_equilibrium=True)

        self.control_volume.add_material_balances(
            balance_type=self.config.material_balance_type, has_phase_equilibrium=True
        )

        self.control_volume.add_energy_balances(
            balance_type=self.config.energy_balance_type,
            has_heat_transfer=self.config.has_heat_transfer,
        )

        self.control_volume.add_momentum_balances(
            balance_type=self.config.momentum_balance_type,
            has_pressure_change=self.config.has_pressure_change,
        )

        # Add Ports
        self.add_inlet_port()

        split_map = {}
        for p in self.control_volume.properties_in.phase_list:
            p_obj = self.control_volume.properties_in.params.get_phase(p)
            if p_obj.is_vapor_phase():
                # Vapor leaves through Vap outlet
                split_map[p] = "Vap"
            else:
                # All other phases leave through Liq outlet
                split_map[p] = "Liq"

        self.split = Separator(
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
            outlet_list=["Vap", "Liq"],
            split_basis=SplittingType.phaseFlow,
            ideal_separation=self.config.ideal_separation,
            ideal_split_map=split_map,
            mixed_state_block=self.control_volume.properties_out,
            has_phase_equilibrium=not self.config.ideal_separation,
            energy_split_basis=self.config.energy_split_basis,
        )
        if not self.config.ideal_separation:

            def split_frac_rule(b, t, o):
                return b.split.split_fraction[t, o, o] == 1

            self.split_fraction_eq = Constraint(
                self.flowsheet().time, self.split.outlet_idx, rule=split_frac_rule
            )

        self.vap_outlet = Port(extends=self.split.Vap)
        self.liq_outlet = Port(extends=self.split.Liq)

        # Add references
        if (
            self.config.has_heat_transfer is True
            and self.config.energy_balance_type != EnergyBalanceType.none
        ):
            self.heat_duty = Reference(self.control_volume.heat[:])
        if (
            self.config.has_pressure_change is True
            and self.config.momentum_balance_type != MomentumBalanceType.none
        ):
            self.deltaP = Reference(self.control_volume.deltaP[:])

    def _get_performance_contents(self, time_point=0):
        var_dict = {}
        if hasattr(self, "heat_duty"):
            var_dict["Heat Duty"] = self.heat_duty[time_point]
        if hasattr(self, "deltaP"):
            var_dict["Pressure Change"] = self.deltaP[time_point]

        return {"vars": var_dict}

    def _get_stream_table_contents(self, time_point=0):
        stream_attributes = {}
        stream_attributes["Units"] = {}

        sblocks = {
            "Inlet": self.control_volume.properties_in,
        }
        if not self.config.ideal_separation:
            # If not using ideal separation, we can get outlet state directly
            # from the state blocks
            sblocks["Vapor Outlet"] = self.split.Vap_state
            sblocks["Liquid Outlet"] = self.split.Liq_state

        for n, v in sblocks.items():
            dvars = v[time_point].define_display_vars()

            stream_attributes[n] = {}

            for k in dvars:
                for i in dvars[k].keys():
                    stream_key = k if i is None else f"{k} {i}"

                    quant = report_quantity(dvars[k][i])

                    stream_attributes[n][stream_key] = quant.m
                    stream_attributes["Units"][stream_key] = quant.u

        if self.config.ideal_separation:
            # If using ideal separation, get values from Ports and hope they map
            # to names in Inlet
            # TODO: Add a better way to map these if necessary
            for n, v in {
                "Vapor Outlet": "vap_outlet",
                "Liquid Outlet": "liq_outlet",
            }.items():
                port_obj = getattr(self, v)

                stream_attributes[n] = {}

                for k in port_obj.vars:
                    for i in port_obj.vars[k].keys():
                        if isinstance(i, float):
                            quant = report_quantity(port_obj.vars[k][time_point])
                            stream_attributes[n][k] = quant.m
                            stream_attributes["Units"][k] = quant.u
                        else:
                            if len(i) == 2:
                                kname = str(i[1])
                            else:
                                kname = str(i[1:])
                            quant = report_quantity(port_obj.vars[k][time_point, i[1:]])
                            stream_attributes[n][k + " " + kname] = quant.m
                            stream_attributes["Units"][k + " " + kname] = quant.u

        return DataFrame.from_dict(stream_attributes, orient="columns")



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for demonstration flowsheet.

"""

import pytest

from idaes.models.flowsheets.demo_flowsheet import (
    build_flowsheet,
    set_dof,
    set_scaling,
    initialize_flowsheet,
    solve_flowsheet,
)

from pyomo.environ import value
from pyomo.network import Arc
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.models.unit_models import Mixer, Heater, Flash
from idaes.core.util.model_statistics import degrees_of_freedom


@pytest.fixture(scope="module")
def model():
    m = build_flowsheet()

    return m


@pytest.mark.unit
def test_build_flowsheet(model):
    assert isinstance(model.fs, FlowsheetBlock)

    assert isinstance(model.fs.BT_props, BTXParameterBlock)

    assert isinstance(model.fs.M01, Mixer)
    assert isinstance(model.fs.H02, Heater)
    assert isinstance(model.fs.F03, Flash)

    assert isinstance(model.fs.s01, Arc)
    assert isinstance(model.fs.s02, Arc)

    assert degrees_of_freedom(model) == 13


@pytest.mark.unit
def test_set_scaling(model):
    set_scaling(model)


@pytest.mark.unit
def test_set_dof(model):
    set_dof(model)

    assert degrees_of_freedom(model) == 0


@pytest.mark.unit
def test_initialize_flowsheet(model):
    initialize_flowsheet(model)

    assert degrees_of_freedom(model) == 0

    assert model.fs.M01.outlet.flow_mol[0].value == pytest.approx(2.0, 1e-3)
    assert model.fs.H02.outlet.flow_mol[0].value == pytest.approx(2.0, 1e-3)
    assert model.fs.F03.vap_outlet.flow_mol[0].expr.value == pytest.approx(1.367, 1e-3)
    assert model.fs.F03.liq_outlet.flow_mol[0].expr.value == pytest.approx(0.633, 1e-3)


@pytest.mark.integration
def test_unit_consistency(model):
    assert_units_consistent(model)


@pytest.mark.unit
def test_solve_flowsheet(model):
    solve_flowsheet(model)

    assert model.fs.M01.outlet.flow_mol[0].value == pytest.approx(2.0, 1e-4)
    assert model.fs.M01.outlet.mole_frac_comp[0, "benzene"].value == pytest.approx(
        0.5, 1e-4
    )
    assert model.fs.M01.outlet.mole_frac_comp[0, "toluene"].value == pytest.approx(
        0.5, 1e-4
    )
    assert model.fs.M01.outlet.pressure[0].value == pytest.approx(101325, 1e-4)
    assert model.fs.M01.outlet.temperature[0].value == pytest.approx(368.12, 1e-4)

    assert model.fs.H02.outlet.flow_mol[0].value == pytest.approx(2.0, 1e-4)
    assert model.fs.H02.outlet.mole_frac_comp[0, "benzene"].value == pytest.approx(
        0.5, 1e-4
    )
    assert model.fs.H02.outlet.mole_frac_comp[0, "toluene"].value == pytest.approx(
        0.5, 1e-4
    )
    assert model.fs.H02.outlet.pressure[0].value == pytest.approx(101325, 1e-4)
    assert model.fs.H02.outlet.temperature[0].value == pytest.approx(370, 1e-4)

    assert value(model.fs.F03.vap_outlet.flow_mol[0]) == pytest.approx(1.3673, 1e-4)
    assert value(model.fs.F03.vap_outlet.mole_frac_comp[0, "benzene"]) == pytest.approx(
        0.5694, 1e-4
    )
    assert value(model.fs.F03.vap_outlet.mole_frac_comp[0, "toluene"]) == pytest.approx(
        0.4306, 1e-4
    )
    assert value(model.fs.F03.vap_outlet.pressure[0]) == pytest.approx(101325, 1e-4)
    assert value(model.fs.F03.vap_outlet.temperature[0]) == pytest.approx(370, 1e-4)

    assert value(model.fs.F03.liq_outlet.flow_mol[0]) == pytest.approx(0.6327, 1e-4)
    assert value(model.fs.F03.liq_outlet.mole_frac_comp[0, "benzene"]) == pytest.approx(
        0.3501, 1e-4
    )
    assert value(model.fs.F03.liq_outlet.mole_frac_comp[0, "toluene"]) == pytest.approx(
        0.6499, 1e-4
    )
    assert value(model.fs.F03.liq_outlet.pressure[0]) == pytest.approx(101325, 1e-4)
    assert value(model.fs.F03.liq_outlet.temperature[0]) == pytest.approx(370, 1e-4)
